from dp5.nmr_processing.nmr_ai import NMRData

from pathlib import Path
import pickle
import logging

from rdkit import Chem

logger = logging.getLogger(__name__)

data = NMRData(
    ["/home/rk582/DP5/test/NMR_data/Compound1/6011"], "chloroform", Path.cwd()
)

print(f"NMRData has processed carbon FID: {hasattr(data, 'carbon_fid')}")

with open("/home/rk582/DP5/test/cmpd1.preint", "wb") as f:
    pickle.dump(data, f)
