from dp5.dft.gaussian import DFTMethod
from dp5.run.data_structures import get
from dp5.run.data_structures import Molecule
from dp5.nmr_processing.nmr_ai import NMRData
import os

s11 = '/home/rk582/DP5/S11'

nmr_data = NMRData(['/home/rk582/DP5/S11/S11_NMR'], 'chloroform')

mol = Molecule('S11/S11b_.sdf')

settings = {'charge': 0, 'optimisation_converged': True,
            'dft_complete': True, 'memory': 1000, 'num_processors': 42, 'executable': {'gaussian': ''}}

os.chdir('/home/rk582/DP5/S11')
g16 = DFTMethod(settings)
g16.nmr([mol])
