from dp5.nmr_processing.nmr_ai import NMRData

from pathlib import Path
import pickle
import logging

from rdkit import Chem

logger = logging.getLogger(__name__)

data = NMRData(["/home/rk582/DP5/test/TPS-1-14/1"],
               "chloroform", Path.cwd())

print(f"NMRData has processed proton FID: {hasattr(data, 'proton_fid')}")

with open('/home/rk582/DP5/test/tps-1-14.preint', 'wb') as f:
    pickle.dump(data, f)
